//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "BandwidthControl.h"
#include "emule.h"
#include "preferences.h"
#include "math.h"

IMPLEMENT_DYNCREATE(CBandwidthControl, CObject)
//Changed by Pawcio
CBandwidthControl::CBandwidthControl(void){
	currUpload				= 0.0f;
    retUpload				= 0.0f;
	minUpReached			= false;
	currDownload			= 0.0f;
    retDownload				= 0.0f;
	maxUpDataRate			= 0.0f;
	maxDownDataRate			= 0.0f;
	eMuleUpDataRate			= 0.0f;
	eMuleUpRate				= 0.0f;
	eMuleoverhead			= 0.0f;
	eMuleStatsUpRate		= 0.0f;
	eMuleStatsDownRate		= 0.0f;

	for (int i=0; i<MAX_AVE_TIMES; i++){
		eMuleUpAvg[i].ave=0;
		eMuleUpAvg[i].Bps=0;
		eMuleUpAvg[i].time=0;
		eMuleUpAvg[i].total=0.0;
		eMuleUpDataAvg[i].ave=0;
		eMuleUpDataAvg[i].Bps=0;
		eMuleUpDataAvg[i].time=0;
		eMuleUpDataAvg[i].total=0.0;
		eMuleUpStats[i].ave=0;
		eMuleUpStats[i].Bps=0;
		eMuleUpStats[i].time=0;
		eMuleUpStats[i].total=0.0;
		eMuleDownStats[i].ave=0;
		eMuleDownStats[i].Bps=0;
		eMuleDownStats[i].time=0;
		eMuleDownStats[i].total=0.0;
	}
	eMuleUp=0;
	eMuleUpData=0;
	eMuleDown=0;
	sessionLimit = false;

	pingThread = NULL;
	pingThreadFailed = false;
	pingUpSliderPos = 0xff;
	pingDownSliderPos = 0xff;
	base_ping_up = 100.0f;
	base_ping_down = 200.0f;

	avg_ping=0;

	useUploadTemp=false;
	useDownloadTemp=false;
	dbUp=0; dbUpData=0; dbDown=0;

	m_dbTotalUp = eMuleUp;
	m_dbTotalUpData = eMuleUpData;
	m_dbTotalDown = eMuleDown;
	m_nArrayIndex = 0;
	upload = 0;

	currUpload=theApp.glob_prefs->GetMinBCUpload();
	currDownload=theApp.glob_prefs->GetMinBCDownload();
}

CBandwidthControl::~CBandwidthControl(void){
	if (pingThread) 
		delete pingThread;
}

//Changed by Pawcio
int CBandwidthControl::Process(){
	dwTime = ::GetTickCount();
	if (!pingThread && !pingThreadFailed && (theApp.glob_prefs->IsUSSEnabled() || theApp.glob_prefs->IsDSSEnabled()))
		StartPinging();
	else if ((pingThread || pingThreadFailed) && !theApp.glob_prefs->IsUSSEnabled() && !theApp.glob_prefs->IsDSSEnabled()){
		StopPinging();
		useUploadTemp=false;
	}

	dbUp = eMuleUp;
	dbUpData = eMuleUpData;
	dbDown = eMuleDown;

	//calc the average bps
	CalcAverages( dbUp, dwTime, 0, &eMuleUpAvg[0],AVE_TIMES);
	CalcAverages( dbUpData, dwTime, 0, &eMuleUpDataAvg[0],AVE_TIMES);
	CalcAverages( dbUp, dwTime, 0, &eMuleUpStats[0],STATS_AVE_TIMES);
	CalcAverages( dbDown, dwTime, 0, &eMuleDownAvg[0],AVE_TIMES);
	CalcAverages( dbDown, dwTime, 0, &eMuleDownStats[0],STATS_AVE_TIMES);

	//save the totals
	m_dbTotalUp=dbUp;
	m_dbTotalUpData=dbUpData;
	m_dbTotalDown=dbDown;

	eMuleUpRateTemp = eMuleUpAvg[m_nArrayIndex].ave/1024.0f;
	eMuleStatsUpRateTemp = eMuleUpStats[m_nArrayIndex].ave/1024.0f;
	eMuleStatsDownRateTemp = eMuleDownStats[m_nArrayIndex].ave/1024.0f;
	eMuleUpDataRate = eMuleUpDataAvg[m_nArrayIndex].ave/1024.0f;
	eMuleDownRate = eMuleDownAvg[m_nArrayIndex].ave/1024.0f;
	eMuleUpRate = eMuleUpRateTemp + 0.025f * eMuleDownRate;
	eMuleStatsDownRate = eMuleStatsDownRateTemp + 0.025f * eMuleStatsUpRateTemp;
	eMuleStatsUpRate = eMuleStatsUpRateTemp + 0.025f * eMuleStatsDownRateTemp;
	eMuleoverhead = eMuleUpRate - eMuleUpDataRate;

	if (eMuleoverhead<0.0f) eMuleoverhead=0.0f;

	if (pingThread && (pingThread->isPreparing(true) || pingThread->isWorking())){
		if (pingThread->isPreparing(true)){
			if (!useUploadTemp){
				useUploadTemp = true;
				uploadTemp = currUpload;
			}
			upload = (theApp.uploadqueue->GetUploadQueueLength() + theApp.uploadqueue->GetReservedQueueCount()) * 0.5f;
			if (upload < 1.0f) upload = 1.0f;

			if (!useDownloadTemp){
				useDownloadTemp = true;
				downloadTemp = currDownload;
			}
			download = theApp.downloadqueue->GetDownloadQueueLength() * 0.5f;
			if (download < 1.0f) download = 1.0f;
		}
		else if (pingThread->isWorking()){
			pingVal ping = pingThread->GetPing();
			if (!theApp.glob_prefs->IsManualUpMaxPing() && (pingUpSliderPos != theApp.glob_prefs->GetUSSSliderPos() || ping.lowest_changed) || !theApp.glob_prefs->IsManualDownMaxPing() && (pingDownSliderPos != theApp.glob_prefs->GetDSSSliderPos() || ping.lowest_changed))
				ReCalculateBasePing(ping.lowest_ping);
			if (theApp.glob_prefs->IsManualUpMaxPing())
				base_ping_up = theApp.glob_prefs->GetBasePingUp();
			if (theApp.glob_prefs->IsManualDownMaxPing())
				base_ping_down = theApp.glob_prefs->GetBasePingDown();
			//Upload --->
			if (theApp.glob_prefs->IsUSSEnabled()) {
				if (base_ping_up < 1.0f)
					base_ping_up = 1.0f;
				//Bandwidth controlled by ping
				if (ping.changed){
					if (useUploadTemp){
						useUploadTemp = false;
						upload = uploadTemp;
					}
					else
						upload = currUpload;

					// ZZ:UploadSpeedSense -->
					float normalized_ping = ping.avg_ping - ping.lowest_ping;
					float hping = base_ping_up - normalized_ping;

					float multiplier = abs(hping / base_ping_up);
					if (multiplier > 1.0f){
						hping /= multiplier;
						if (multiplier > 2.0f)
							hping *= (log10f(multiplier)/0.301f);
					}

					// Calculate change of upload speed
					float ulDiff = hping / (ping.lowest_ping*100.0f);
					float additionalUp = (UPLOAD_CLIENT_DATARATE + eMuleUpRate*20.5f)/1024.0f;
					if (additionalUp > 10.0f)
						additionalUp = 10.0f;
					if (ulDiff < 0 || eMuleUpRate + additionalUp > upload) //Don't increase when upload level is not reached
						upload += ulDiff;
					// ZZ:UploadSpeedSense <--
				}
			}
			//Upload <---

			//Download --->
			if (theApp.glob_prefs->IsDSSEnabled()) {
				if (base_ping_down < 1.0f)
					base_ping_down = 1.0f;
				//Bandwidth controlled by ping
				if (ping.changed){
					if (useDownloadTemp){
						useDownloadTemp = false;
						download = downloadTemp;
					}
					else
						download = currDownload;

					// DownloadSpeedSense -->
					float normalized_ping = ping.avg_ping - ping.lowest_ping;
					float hping = base_ping_down - normalized_ping;

					float multiplier = abs(hping / base_ping_down);
					if (multiplier > 1.0f){
						hping /= multiplier;
						if (multiplier > 2.0f)
							hping *= (log10f(multiplier)/0.2f);
					}

					// Calculate change of upload speed
					float downDiff = hping / (ping.lowest_ping * 64.0f);
					float additionalDown = (UPLOAD_CLIENT_DATARATE + maxDownDataRate*100.0f)/1024.0f;
					if (additionalDown > 10.0f)
						additionalDown = 10.0f;
					if (downDiff < 0 || eMuleDownRate + additionalDown > download) //Don't increase when download level is not reached
						download += downDiff;
					// DownloadSpeedSense <--
				}
			}
			//Download <---
			avg_ping = ping.avg_ping;
		}
		if (theApp.glob_prefs->IsUSSEnabled()){
			// Do not fall under the minUpload allowed
			if(upload < theApp.glob_prefs->GetMinBCUpload() && !useUploadTemp){
				minUpReached = true;
				upload = theApp.glob_prefs->GetMinBCUpload();
			}
			else {
				minUpReached = false;
				if (upload > theApp.glob_prefs->GetMaxBCUpload() && !theApp.glob_prefs->IsAutoMaxUpload())
					upload = theApp.glob_prefs->GetMaxBCUpload();
			}
		}
		if (theApp.glob_prefs->IsDSSEnabled()){
			// Do not fall under the minDownload allowed
			if(download < theApp.glob_prefs->GetMinBCDownload() && !useDownloadTemp)
				download = theApp.glob_prefs->GetMinBCDownload();
			else if (download > theApp.glob_prefs->GetMaxBCDownload() && !theApp.glob_prefs->IsAutoMaxDownload())
				download = theApp.glob_prefs->GetMaxBCDownload();
		}
	}
	if (!theApp.glob_prefs->IsUSSEnabled() || (pingThread && !pingThread->isPreparing(true) && !pingThread->isWorking())){
		minUpReached = true;
		upload=theApp.glob_prefs->GetMaxUpload();
	}
	if (!theApp.glob_prefs->IsDSSEnabled() || (pingThread && !pingThread->isPreparing(true) && !pingThread->isWorking()))
		download=theApp.glob_prefs->GetMaxDownload();

	uploadACK = currUpload = upload;
	if (theApp.glob_prefs->IsIncludeTCPAck()){
		uploadACK -= (0.025f * eMuleDownRate);
		if (uploadACK < 1.0f)
			uploadACK = 1.0f;
	}
	if (theApp.uploadqueue->GetDatarate() > maxUpDataRate * 1024.0f)
		maxUpDataRate = theApp.uploadqueue->GetDatarate() / 1024.0f;
	retUpload = uploadACK;

	downloadACK = currDownload = download;
	if (theApp.glob_prefs->IsIncludeTCPAck() && download != UNLIMITED){
		downloadACK -= (0.025f * eMuleUpRate);
		if (downloadACK < 1.0f)
			downloadACK = 1.0f;
	}
	if (theApp.downloadqueue->GetDatarate() > maxDownDataRate * 1024.0f)
		maxDownDataRate = theApp.downloadqueue->GetDatarate() / 1024.0f;
	retDownload = downloadACK;

	//increment the circular buffer index
	if( ++m_nArrayIndex >= MAX_AVE_TIMES  )
		m_nArrayIndex = 0;
	return 0;
}

void CBandwidthControl::ReCalculateBasePing(float lowestPing){
	pingUpSliderPos = theApp.glob_prefs->GetUSSSliderPos();

	if (lowestPing < 0.1f) 
		lowestPing = 0.1f;
	else if (lowestPing > 2000.0f)
		lowestPing = 2000.0f;

	float bestPingUp = lowestPing * 13.65f / (log10f(lowestPing + 3)*log10f(lowestPing + 4));
	float multiplier = 4.0f / log10f(8 + lowestPing);
	if (pingUpSliderPos < 12)
		base_ping_up = bestPingUp - ((bestPingUp - (bestPingUp/multiplier)) * (12.0f - pingUpSliderPos)/11.0f);
	else if (pingUpSliderPos > 12)
		base_ping_up = bestPingUp + (((bestPingUp*multiplier) - bestPingUp) * (pingUpSliderPos - 12.0f)/12.0f);
	else
		base_ping_up=bestPingUp;

	pingDownSliderPos = theApp.glob_prefs->GetDSSSliderPos();

	float bestPingDown = bestPingUp * (multiplier/3.0f + 1.0f);
	if (pingDownSliderPos < 12)
		base_ping_down = bestPingDown - ((bestPingDown - (bestPingDown/multiplier)) * (12.0f - pingDownSliderPos)/11.0f);
	else if (pingDownSliderPos > 12)
		base_ping_down = bestPingDown + (((bestPingDown*multiplier) - bestPingDown) * (pingDownSliderPos - 12.0f)/12.0f);
	else
		base_ping_down=bestPingDown;
}

float CBandwidthControl::GetMaxDownloadOnUpload(float upload, float maxdownload, float isConnectionOH){
	if( upload < 4.0f + isConnectionOH / 2.0f )
		return (( (upload < 10.0f + isConnectionOH) && (upload*3.0f < maxdownload) )? upload*3.0f : maxdownload);
	return (( (upload < 10.0f + isConnectionOH) && (upload*4.0f < maxdownload) )? upload*4.0f : maxdownload);
}

float CBandwidthControl::GetMaxDownload(){
	float isConnectionOH = (float) (theApp.glob_prefs->IsConnectionsOverHead() && theApp.glob_prefs->IsManageConnection());

	if (theApp.glob_prefs->IsZZRatio() || theApp.glob_prefs->IsIncludeTCPAck() || (theApp.glob_prefs->IsConnectionsOverHead() && !theApp.glob_prefs->IsManageConnection())){
		//MORPH START - Changed by IceCream, Increase the ZZ upload:download ratio from 1:3 to 1:4 [IceCream]
		if(theApp.stat_sessionReceivedBytes > theApp.stat_sessionSentBytes * 4 && eMuleUpDataRate < 10.0f){
			//Pawcio: Moved from DownloadQueue::Process		
			float tempDownspeed = 4.0f * (float)theApp.uploadqueue->GetDatarate()/1024 * 16.0f * pow((float)theApp.stat_sessionSentBytes/theApp.stat_sessionReceivedBytes, 2);
			if(tempDownspeed < retDownload){
				sessionLimit = true;
				return tempDownspeed;
			}
			//End Pawcio	
		}
		//MORPH END   - Changed by IceCream, Increase the ZZ upload:download ratio from 1:3 to 1:4 [IceCream]
		sessionLimit = false;
		return retDownload;
	}
	sessionLimit = false;
	return GetMaxDownloadOnUpload(currUpload, retDownload, isConnectionOH);
}

void CBandwidthControl::AddUp(uint32 up, bool fileData){
	if (fileData) {
		eMuleUpData+=(up-20-20);
	}
	eMuleUp+=up;
}

void CBandwidthControl::AddDown(uint32 down, bool fileData){
	eMuleDown+=down;
}

void CBandwidthControl::StartBandwidthControl(){
	if(Init())
		theApp.emuledlg->AddLogLine(false,"BandwidthControl V1.0 initialized - made by Pawcio (based on AMUC by Herbert / DynamicUP by Tarod, Hardware / UploadSpeedSense by ZZ).");
}

bool CBandwidthControl::Init(){
	// start ping thread
	if (theApp.glob_prefs->IsUSSEnabled() || theApp.glob_prefs->IsDSSEnabled()){
		pingThread = (CPingThread*)AfxBeginThread(RUNTIME_CLASS(CPingThread), THREAD_PRIORITY_HIGHEST, 0, CREATE_SUSPENDED);
		if (pingThread){
			pingThread->Init();
			pingThread->StartPingThread();
		}
		else
			theApp.emuledlg->QueueDebugLogLine(false, "Can't initialize Speed Sense! USS and DSS are not able to work!");
	}
	return true;
}

void CBandwidthControl::CalcAverages( uint64 dbTotal, DWORD dwTime, DWORD dwBps, STATS_STRUCT* pStats, int g_nAveraging )
{
    ASSERT( g_nAveraging <= MAX_AVE_TIMES );

    //set current bps
    pStats[m_nArrayIndex].Bps = dwBps;

    //set total bytes sent or recv
    pStats[m_nArrayIndex].total = dbTotal;

    //set the current time (milliseconds)
    pStats[m_nArrayIndex].time = dwTime;

    //start is the index in the array for calculating averages
    int start = m_nArrayIndex - g_nAveraging;
	if( start < 0 )
		start = MAX_AVE_TIMES + start;

    //the array entry may not have been filled in yet
    if( pStats[start].time == 0 ) //Pawcio
        start = 0;

    //set average based upon sampling window size
	uint64 dbSampleTotal = 0;

    //total bytes received/sent in our sampling window
	if( pStats[start].time ) {
		if (dbTotal >= pStats[start].total)
			dbSampleTotal = dbTotal - pStats[start].total;
		else {
			dbSampleTotal = 0xffffffff - pStats[start].total;
			dbSampleTotal += dbTotal;
		}
	}

    //elapsed milliseconds
    DWORD dwElapsed = ( pStats[m_nArrayIndex].time - pStats[start].time );

    //calc average
    if( dwElapsed )
        pStats[m_nArrayIndex].ave = dbSampleTotal*1000i64/dwElapsed;
}

void	CBandwidthControl::StartPinging(){
	if (pingThread)
		return;

	pingThread = (CPingThread*)AfxBeginThread(RUNTIME_CLASS(CPingThread), THREAD_PRIORITY_HIGHEST, 0, CREATE_SUSPENDED);
	if (pingThread){
		pingThread->Init();
		pingThread->StartPingThread();
	}
	else {
		pingThreadFailed = true;
		theApp.emuledlg->QueueDebugLogLine(false, "Can't initialize Speed Sense! USS and DSS are not able to work!");
	}

}

void CBandwidthControl::StopPinging(){
	pingThreadFailed = false;
	if (!pingThread)
		return;
	pingThread->EndPingThread();
	delete pingThread;
	pingThread=NULL;
}


// Finish
void CBandwidthControl::EndBandwidthControl(void)
{
	if (pingThread)
		pingThread->EndPingThread();
}

bool CBandwidthControl::SetIP(CString IP){
	bool retVal=true;
	if (pingThread){
		pingThread->StopPingThread();
		if (!pingThread->SetIP(IP))
			retVal=false;
		pingThread->StartPingThread();
	}
	return retVal;
}
